@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2022 Jim Ramsay
+ * Copyright (c) 2023 Andri Yngvason
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -18,6 +19,8 @@
 
 #include "output.h"
 
+#include <sys/socket.h>
+
 struct ctl;
 struct cmd_response;
 
@@ -25,9 +28,12 @@ struct ctl_server_client;
 
 struct ctl_server_client_info {
 	int id;
-	const char *hostname;
-	const char *username;
-	const char *seat;
+	union {
+		struct sockaddr_storage address_storage;
+		struct sockaddr address;
+	};
+	const char* username;
+	const char* seat;
 };
 
 struct ctl_server_output {
@@ -41,6 +47,8 @@ struct ctl_server_output {
 
 struct ctl_server_actions {
 	void* userdata;
+	struct cmd_response* (*on_attach)(struct ctl*, const char* display);
+	struct cmd_response* (*on_detach)(struct ctl*);
 	struct cmd_response* (*on_output_cycle)(struct ctl*,
 			enum output_cycle_direction direction);
 	struct cmd_response* (*on_output_switch)(struct ctl*,
@@ -79,3 +87,5 @@ void ctl_server_event_disconnected(struct ctl*,
 
 void ctl_server_event_capture_changed(struct ctl*,
 		const char* captured_output);
+
+void ctl_server_event_detached(struct ctl*);
