@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2022-2023 Jim Ramsay
+ * Copyright (c) 2023 Andri Yngvason
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -26,6 +27,7 @@
 #include <neatvnc.h>
 #include <aml.h>
 #include <jansson.h>
+#include <arpa/inet.h>
 
 #include "output.h"
 #include "ctl-commands.h"
@@ -46,6 +48,11 @@ struct cmd {
 	enum cmd_type type;
 };
 
+struct cmd_attach {
+	struct cmd cmd;
+	char display[128];
+};
+
 struct cmd_help {
 	struct cmd cmd;
 	char id[64];
@@ -105,6 +112,19 @@ static void cmd_response_destroy(struct cmd_response* self)
 	free(self);
 }
 
+static struct cmd_attach* cmd_attach_new(json_t* args,
+		struct jsonipc_error* err)
+{
+	const char* display = NULL;
+	if (json_unpack(args, "{s:s}", "display", &display) == -1) {
+		jsonipc_error_printf(err, EINVAL, "Missing display name");
+		return NULL;
+	}
+	struct cmd_attach* cmd = calloc(1, sizeof(*cmd));
+	strlcpy(cmd->display, display, sizeof(cmd->display));
+	return cmd;
+}
+
 static struct cmd_help* cmd_help_new(json_t* args,
 		struct jsonipc_error* err)
 {
@@ -185,6 +205,9 @@ static struct cmd* parse_command(struct jsonipc_request* ipc,
 	enum cmd_type cmd_type = ctl_command_parse_name(ipc->method);
 	struct cmd* cmd = NULL;
 	switch (cmd_type) {
+	case CMD_ATTACH:
+		cmd = (struct cmd*)cmd_attach_new(ipc->params, err);
+		break;
 	case CMD_HELP:
 		cmd = (struct cmd*)cmd_help_new(ipc->params, err);
 		break;
@@ -194,6 +217,7 @@ static struct cmd* parse_command(struct jsonipc_request* ipc,
 	case CMD_CLIENT_DISCONNECT:
 		cmd = (struct cmd*)cmd_disconnect_client_new(ipc->params, err);
 		break;
+	case CMD_DETACH:
 	case CMD_VERSION:
 	case CMD_EVENT_RECEIVE:
 	case CMD_CLIENT_LIST:
@@ -339,6 +363,25 @@ static struct ctl_server_client* ctl_server_client_next(struct ctl* self,
 	return self->actions.client_next(self, prev);
 }
 
+static int sockaddr_to_string(char* dst, size_t sz, const struct sockaddr* addr)
+{
+	struct sockaddr_in* sa_in = (struct sockaddr_in*)addr;
+	struct sockaddr_in6* sa_in6 = (struct sockaddr_in6*)addr;
+
+	switch (addr->sa_family) {
+	case AF_INET:
+		inet_ntop(addr->sa_family, &sa_in->sin_addr, dst, sz);
+		return 0;
+	case AF_INET6:
+		inet_ntop(addr->sa_family, &sa_in6->sin6_addr, dst, sz);
+		return 0;
+	}
+	nvnc_log(NVNC_LOG_DEBUG,
+			"Don't know how to convert sa_family %d to string",
+			addr->sa_family);
+	return -1;
+}
+
 static void ctl_server_client_get_info(struct ctl* self,
 		const struct ctl_server_client* client,
 		struct ctl_server_client_info* info)
@@ -361,9 +404,12 @@ static struct cmd_response* generate_vnc_client_list(struct ctl* self)
 		snprintf(id_str, sizeof(id_str), "%d", info.id);
 		json_t* packed = json_pack("{s:s}", "id", id_str);
 
-		if (info.hostname)
-			json_object_set_new(packed, "hostname",
-					json_string(info.hostname));
+		char address_string[256];
+		if (sockaddr_to_string(address_string, sizeof(address_string),
+					&info.address) == 0) {
+			json_object_set_new(packed, "address",
+					json_string(address_string));
+		}
 
 		if (info.username)
 			json_object_set_new(packed, "username",
@@ -407,6 +453,11 @@ static struct cmd_response* ctl_server_dispatch_cmd(struct ctl* self,
 	nvnc_log(NVNC_LOG_INFO, "Dispatching control client command '%s'", info->name);
 	struct cmd_response* response = NULL;
 	switch (cmd->type) {
+	case CMD_ATTACH:{
+		struct cmd_attach* c = (struct cmd_attach*)cmd;
+		response = self->actions.on_attach(self, c->display);
+		break;
+		}
 	case CMD_HELP:{
 		struct cmd_help* c = (struct cmd_help*)cmd;
 		response = generate_help_object(c->id, c->id_is_command);
@@ -423,6 +474,9 @@ static struct cmd_response* ctl_server_dispatch_cmd(struct ctl* self,
 		response = self->actions.on_disconnect_client(self, c->id);
 		break;
 		}
+	case CMD_DETACH:
+		response = self->actions.on_detach(self);
+		break;
 	case CMD_WAYVNC_EXIT:
 		response = self->actions.on_wayvnc_exit(self);
 		break;
@@ -866,9 +920,13 @@ json_t* pack_connection_event_params(
 	char id_str[64];
 	snprintf(id_str, sizeof(id_str), "%d", info->id);
 
+	char address_string[256];
+	bool have_addr = sockaddr_to_string(address_string,
+			sizeof(address_string), &info->address) == 0;
+
 	return json_pack("{s:s, s:s?, s:s?, s:s?, s:i}",
 			"id", id_str,
-			"hostname", info->hostname,
+			"address", have_addr ? address_string : NULL,
 			"username", info->username,
 			"seat", info->seat,
 			"connection_count", new_connection_count);
@@ -942,3 +1000,8 @@ void ctl_server_event_capture_changed(struct ctl* self,
 	ctl_server_enqueue_event(self, EVT_CAPTURE_CHANGED,
 			json_pack("{s:s}", "output", captured_output));
 }
+
+void ctl_server_event_detached(struct ctl* self)
+{
+	ctl_server_enqueue_event(self, EVT_DETACHED, json_object());
+}
