@@ -39,11 +39,14 @@
 #include "virtual-keyboard-unstable-v1.h"
 #include "xdg-output-unstable-v1.h"
 #include "wlr-output-power-management-unstable-v1.h"
+#include "wlr-output-management-unstable-v1.h"
 #include "linux-dmabuf-unstable-v1.h"
+#include "ext-transient-seat-v1.h"
 #include "screencopy.h"
 #include "data-control.h"
 #include "strlcpy.h"
 #include "output.h"
+#include "output-management.h"
 #include "pointer.h"
 #include "keyboard.h"
 #include "seat.h"
@@ -71,6 +74,8 @@
 
 #define MAYBE_UNUSED __attribute__((unused))
 
+struct wayvnc_client;
+
 enum socket_type {
 	SOCKET_TYPE_TCP = 0,
 	SOCKET_TYPE_UNIX,
@@ -82,6 +87,7 @@ struct wayvnc {
 
 	struct wl_display* display;
 	struct wl_registry* registry;
+	struct aml_handler* wl_handler;
 	struct wl_list outputs;
 	struct wl_list seats;
 	struct cfg cfg;
@@ -89,6 +95,7 @@ struct wayvnc {
 	struct zwp_virtual_keyboard_manager_v1* keyboard_manager;
 	struct zwlr_virtual_pointer_manager_v1* pointer_manager;
 	struct zwlr_data_control_manager_v1* data_control_manager;
+	struct ext_transient_seat_manager_v1* transient_seat_manager;
 
 	struct output* selected_output;
 	struct seat* selected_seat;
@@ -108,6 +115,7 @@ struct wayvnc {
 	uint32_t n_frames_captured;
 
 	bool disable_input;
+	bool use_transient_seat;
 
 	int nr_clients;
 	struct aml_ticker* performance_ticker;
@@ -116,6 +124,10 @@ struct wayvnc {
 
 	struct ctl* ctl;
 	bool is_initializing;
+
+	bool start_detached;
+
+	struct wayvnc_client* master_layout_client;
 };
 
 struct wayvnc_client {
@@ -123,6 +135,7 @@ struct wayvnc_client {
 	struct nvnc_client* nvnc_client;
 
 	struct seat* seat;
+	struct ext_transient_seat_v1* transient_seat;
 
 	unsigned id;
 	struct pointer pointer;
@@ -140,6 +153,11 @@ static void client_init_seat(struct wayvnc_client* self);
 static void client_init_pointer(struct wayvnc_client* self);
 static void client_init_keyboard(struct wayvnc_client* self);
 static void client_init_data_control(struct wayvnc_client* self);
+static void client_detach_wayland(struct wayvnc_client* self);
+static int blank_screen(struct wayvnc* self);
+static bool wayland_attach(struct wayvnc* self, const char* display,
+		const char* output);
+static void wayland_detach(struct wayvnc* self);
 
 struct wl_shm* wl_shm = NULL;
 struct zwp_linux_dmabuf_v1* zwp_linux_dmabuf = NULL;
@@ -192,6 +210,12 @@ static bool registry_add_input(void* data, struct wl_registry* registry,
 		return true;
 	}
 
+	if (strcmp(interface, ext_transient_seat_manager_v1_interface.name) == 0) {
+		self->transient_seat_manager = wl_registry_bind(registry, id,
+				&ext_transient_seat_manager_v1_interface, 1);
+		return true;
+	}
+
 	return false;
 }
 
@@ -238,6 +262,15 @@ static void registry_add(void* data, struct wl_registry* registry,
 		return;
 	}
 
+	if (strcmp(interface, zwlr_output_manager_v1_interface.name) == 0) {
+		nvnc_trace("Registering new wlr_output_manager");
+		struct zwlr_output_manager_v1* wlr_output_manager =
+			wl_registry_bind(registry, id,
+				&zwlr_output_manager_v1_interface, 1);
+		wlr_output_manager_setup(wlr_output_manager);
+		return;
+	}
+
 	if (strcmp(interface, zwlr_screencopy_manager_v1_interface.name) == 0) {
 		self->screencopy.manager =
 			wl_registry_bind(registry, id,
@@ -293,8 +326,13 @@ static void registry_remove(void* data, struct wl_registry* registry,
 		output_destroy(out);
 
 		if (out == self->selected_output) {
-			nvnc_log(NVNC_LOG_ERROR, "No fallback outputs left. Exiting...");
-			wayvnc_exit(self);
+			if (self->start_detached) {
+				nvnc_log(NVNC_LOG_WARNING, "No fallback outputs left. Detaching...");
+				wayland_detach(self);
+			} else {
+				nvnc_log(NVNC_LOG_ERROR, "No fallback outputs left. Exiting...");
+				wayvnc_exit(self);
+			}
 		}
 
 		return;
@@ -350,44 +388,125 @@ static int init_render_node(int* fd)
 }
 #endif
 
-void wayvnc_destroy(struct wayvnc* self)
+static void wayland_detach(struct wayvnc* self)
 {
-	cfg_destroy(&self->cfg);
+	if (!self->display)
+		return;
+
+	aml_stop(aml_get_default(), self->wl_handler);
+	aml_unref(self->wl_handler);
+	self->wl_handler = NULL;
+
+	// Screen blanking is required to release wl_shm of linux_dmabuf.
+	if (self->nvnc)
+		blank_screen(self);
+
+	if (self->nvnc) {
+		struct nvnc_client* nvnc_client;
+		for (nvnc_client = nvnc_client_first(self->nvnc); nvnc_client;
+				nvnc_client = nvnc_client_next(nvnc_client)) {
+			struct wayvnc_client* client =
+				nvnc_get_userdata(nvnc_client);
+			client_detach_wayland(client);
+		}
+	}
+
+	self->selected_output = NULL;
+	self->screencopy.wl_output = NULL;
 
 	output_list_destroy(&self->outputs);
 	seat_list_destroy(&self->seats);
 
+	if (zwp_linux_dmabuf)
+		zwp_linux_dmabuf_v1_destroy(zwp_linux_dmabuf);
+	zwp_linux_dmabuf = NULL;
+
+	if (self->screencopy.manager) {
+		screencopy_stop(&self->screencopy);
+		screencopy_destroy(&self->screencopy);
+		zwlr_screencopy_manager_v1_destroy(self->screencopy.manager);
+	}
+	self->screencopy.manager = NULL;
+
 	if (xdg_output_manager)
 		zxdg_output_manager_v1_destroy(xdg_output_manager);
+	xdg_output_manager = NULL;
 
 	if (wlr_output_power_manager)
 		zwlr_output_power_manager_v1_destroy(wlr_output_power_manager);
+	wlr_output_power_manager = NULL;
+
+	wlr_output_manager_destroy();
 
 	wl_shm_destroy(wl_shm);
+	wl_shm = NULL;
 
 	if (self->keyboard_manager)
 		zwp_virtual_keyboard_manager_v1_destroy(self->keyboard_manager);
+	self->keyboard_manager = NULL;
 
 	if (self->pointer_manager)
 		zwlr_virtual_pointer_manager_v1_destroy(self->pointer_manager);
+	self->pointer_manager = NULL;
 
 	if (self->data_control_manager)
 		zwlr_data_control_manager_v1_destroy(self->data_control_manager);
+	self->data_control_manager = NULL;
 
-	if (self->screencopy.manager)
-		zwlr_screencopy_manager_v1_destroy(self->screencopy.manager);
-
-	if (self->performance_ticker)
+	if (self->performance_ticker) {
+		aml_stop(aml_get_default(), self->performance_ticker);
 		aml_unref(self->performance_ticker);
+	}
+	self->performance_ticker = NULL;
 
 	if (self->capture_retry_timer)
 		aml_unref(self->capture_retry_timer);
+	self->capture_retry_timer = NULL;
 
 	wl_registry_destroy(self->registry);
+	self->registry = NULL;
+
 	wl_display_disconnect(self->display);
+	self->display = NULL;
+
+	if (self->ctl)
+		ctl_server_event_detached(self->ctl);
 }
 
-static int init_wayland(struct wayvnc* self)
+void wayvnc_destroy(struct wayvnc* self)
+{
+	cfg_destroy(&self->cfg);
+	wayland_detach(self);
+}
+
+void on_wayland_event(void* obj)
+{
+	struct wayvnc* self = aml_get_userdata(obj);
+
+	int rc MAYBE_UNUSED = wl_display_prepare_read(self->display);
+	assert(rc == 0);
+
+	if (wl_display_read_events(self->display) < 0) {
+		if (errno == EPIPE || errno == ECONNRESET) {
+			nvnc_log(NVNC_LOG_ERROR, "Compositor has gone away. Exiting...");
+			if (self->start_detached)
+				wayland_detach(self);
+			else
+				wayvnc_exit(self);
+			return;
+		} else {
+			nvnc_log(NVNC_LOG_ERROR, "Failed to read wayland events: %m");
+		}
+	}
+
+	if (wl_display_dispatch_pending(self->display) < 0) {
+		nvnc_log(NVNC_LOG_ERROR, "Failed to dispatch pending");
+		wayland_detach(self);
+		// TODO: Re-attach
+	}
+}
+
+static int init_wayland(struct wayvnc* self, const char* display)
 {
 	self->is_initializing = true;
 	static const struct wl_registry_listener registry_listener = {
@@ -395,9 +514,10 @@ static int init_wayland(struct wayvnc* self)
 		.global_remove = registry_remove,
 	};
 
-	self->display = wl_display_connect(NULL);
+	self->display = wl_display_connect(display);
 	if (!self->display) {
-		const char* display_name = getenv("WAYLAND_DISPLAY");
+		const char* display_name = display ? display:
+			getenv("WAYLAND_DISPLAY");
 		if (!display_name) {
 			nvnc_log(NVNC_LOG_ERROR, "WAYLAND_DISPLAY is not set in the environment");
 		} else {
@@ -439,36 +559,35 @@ static int init_wayland(struct wayvnc* self)
 		goto failure;
 	}
 
+	if (!self->transient_seat_manager && self->use_transient_seat) {
+		nvnc_log(NVNC_LOG_ERROR, "Transient seat protocol not supported by compositor");
+		goto failure;
+	}
+
 	self->screencopy.on_done = on_capture_done;
 	self->screencopy.userdata = self;
 
+	self->wl_handler = aml_handler_new(wl_display_get_fd(self->display),
+	                             on_wayland_event, self, NULL);
+	if (!self->wl_handler)
+		goto failure;
+
+	int rc = aml_start(aml_get_default(), self->wl_handler);
+	if (rc < 0)
+		goto handler_failure;
+
 	return 0;
 
 failure:
 	wl_display_disconnect(self->display);
+	self->display = NULL;
+handler_failure:
+	if (self->wl_handler)
+		aml_unref(self->wl_handler);
+	self->wl_handler = NULL;
 	return -1;
 }
 
-void on_wayland_event(void* obj)
-{
-	struct wayvnc* self = aml_get_userdata(obj);
-
-	int rc MAYBE_UNUSED = wl_display_prepare_read(self->display);
-	assert(rc == 0);
-
-	if (wl_display_read_events(self->display) < 0) {
-		if (errno == EPIPE || errno == ECONNRESET) {
-			nvnc_log(NVNC_LOG_ERROR, "Compositor has gone away. Exiting...");
-			wayvnc_exit(self);
-		} else {
-			nvnc_log(NVNC_LOG_ERROR, "Failed to read wayland events: %m");
-		}
-	}
-
-	if (wl_display_dispatch_pending(self->display) < 0)
-		nvnc_log(NVNC_LOG_ERROR, "Failed to dispatch pending");
-}
-
 void wayvnc_exit(struct wayvnc* self)
 {
 	self->do_exit = true;
@@ -517,17 +636,24 @@ static struct ctl_server_client *client_next(struct ctl* ctl,
 		(struct ctl_server_client*)nvnc_client_first(self->nvnc);
 }
 
+static void compose_client_info(const struct wayvnc_client* client,
+		struct ctl_server_client_info* info)
+{
+	info->id = client->id;
+	socklen_t addrlen = sizeof(info->address);
+	nvnc_client_get_address(client->nvnc_client,
+			(struct sockaddr*)&info->address, &addrlen);
+	info->username = nvnc_client_get_auth_username(client->nvnc_client);
+	info->seat = client->seat ? client->seat->name : NULL;
+}
+
 static void client_info(const struct ctl_server_client* client_handle,
 		struct ctl_server_client_info* info)
 {
 	const struct nvnc_client *vnc_client =
 		(const struct nvnc_client*)client_handle;
 	const struct wayvnc_client *client = nvnc_get_userdata(vnc_client);
-
-	info->id = client->id;
-	info->hostname = nvnc_client_get_hostname(vnc_client);
-	info->username = nvnc_client_get_auth_username(vnc_client);
-	info->seat = client->seat ? client->seat->name : NULL;
+	compose_client_info(client, info);
 }
 
 static int get_output_list(struct ctl* ctl,
@@ -591,23 +717,12 @@ int init_main_loop(struct wayvnc* self)
 {
 	struct aml* loop = aml_get_default();
 
-	struct aml_handler* wl_handler;
-	wl_handler = aml_handler_new(wl_display_get_fd(self->display),
-	                             on_wayland_event, self, NULL);
-	if (!wl_handler)
-		return -1;
-
-	int rc = aml_start(loop, wl_handler);
-	aml_unref(wl_handler);
-	if (rc < 0)
-		return -1;
-
 	struct aml_signal* sig;
 	sig = aml_signal_new(SIGINT, on_signal, self, NULL);
 	if (!sig)
 		return -1;
 
-	rc = aml_start(loop, sig);
+	int rc = aml_start(loop, sig);
 	aml_unref(sig);
 	if (rc < 0)
 		return -1;
@@ -621,6 +736,10 @@ static void on_pointer_event(struct nvnc_client* client, uint16_t x, uint16_t y,
 	struct wayvnc_client* wv_client = nvnc_get_userdata(client);
 	struct wayvnc* wayvnc = wv_client->server;
 
+	if (!wv_client->pointer.pointer) {
+		return;
+	}
+
 	uint32_t xfx = 0, xfy = 0;
 	output_transform_coord(wayvnc->selected_output, x, y, &xfx, &xfy);
 
@@ -631,6 +750,9 @@ static void on_key_event(struct nvnc_client* client, uint32_t symbol,
                          bool is_pressed)
 {
 	struct wayvnc_client* wv_client = nvnc_get_userdata(client);
+	if (!wv_client->keyboard.virtual_keyboard) {
+		return;
+	}
 
 	keyboard_feed(&wv_client->keyboard, symbol, is_pressed);
 }
@@ -639,6 +761,9 @@ static void on_key_code_event(struct nvnc_client* client, uint32_t code,
 		bool is_pressed)
 {
 	struct wayvnc_client* wv_client = nvnc_get_userdata(client);
+	if (!wv_client->keyboard.virtual_keyboard) {
+		return;
+	}
 
 	keyboard_feed_code(&wv_client->keyboard, code + 8, is_pressed);
 }
@@ -653,6 +778,32 @@ static void on_client_cut_text(struct nvnc_client* nvnc_client,
 	}
 }
 
+static bool on_client_resize(struct nvnc_client* nvnc_client,
+		const struct nvnc_desktop_layout* layout)
+{
+	struct wayvnc_client* client = nvnc_get_userdata(nvnc_client);
+	struct wayvnc* self = client->server;
+
+	uint16_t width = nvnc_desktop_layout_get_width(layout);
+	uint16_t height = nvnc_desktop_layout_get_height(layout);
+	struct output* output = client->server->selected_output;
+
+	if (output == NULL)
+		return false;
+
+	if (self->master_layout_client && self->master_layout_client != client)
+		return false;
+
+	self->master_layout_client = client;
+
+	nvnc_log(NVNC_LOG_DEBUG,
+		"Client resolution changed: %ux%u, capturing output %s which is headless: %s",
+		width, height, output->name,
+		output->is_headless ? "yes" : "no");
+
+	return wlr_output_manager_resize_output(output, width, height);
+}
+
 bool on_auth(const char* username, const char* password, void* ud)
 {
 	struct wayvnc* self = ud;
@@ -686,9 +837,15 @@ static struct nvnc_fb* create_placeholder_buffer(uint16_t width, uint16_t height
 
 static int blank_screen(struct wayvnc* self)
 {
-	struct nvnc_fb* placeholder_fb =
-		create_placeholder_buffer(self->selected_output->width,
-				self->selected_output->height);
+	int width = 1280;
+	int height = 720;
+
+	if (self->selected_output) {
+		width = output_get_transformed_width(self->selected_output);
+		height = output_get_transformed_height(self->selected_output);
+	}
+
+	struct nvnc_fb* placeholder_fb = create_placeholder_buffer(width, height);
 	if (!placeholder_fb) {
 		nvnc_log(NVNC_LOG_ERROR, "Failed to allocate a placeholder buffer");
 		return -1;
@@ -751,6 +908,8 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 
 	nvnc_set_name(self->nvnc, "WayVNC");
 
+	nvnc_set_desktop_layout_fn(self->nvnc, on_client_resize);
+
 	enum nvnc_auth_flags auth_flags = 0;
 	if (self->cfg.enable_auth) {
 		auth_flags |= NVNC_AUTH_REQUIRE_AUTH;
@@ -792,13 +951,10 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 		}
 	}
 
-	if (self->pointer_manager)
-		nvnc_set_pointer_fn(self->nvnc, on_pointer_event);
+	nvnc_set_pointer_fn(self->nvnc, on_pointer_event);
 
-	if (self->keyboard_manager) {
-		nvnc_set_key_fn(self->nvnc, on_key_event);
-		nvnc_set_key_code_fn(self->nvnc, on_key_code_event);
-	}
+	nvnc_set_key_fn(self->nvnc, on_key_event);
+	nvnc_set_key_code_fn(self->nvnc, on_key_code_event);
 
 	nvnc_set_new_client_fn(self->nvnc, on_nvnc_client_new);
 	nvnc_set_cut_text_fn(self->nvnc, on_client_cut_text);
@@ -1061,6 +1217,34 @@ static void stop_performance_ticker(struct wayvnc* self)
 	aml_stop(aml_get_default(), self->performance_ticker);
 }
 
+static void client_init_wayland(struct wayvnc_client* self)
+{
+	client_init_seat(self);
+	client_init_keyboard(self);
+	client_init_pointer(self);
+	client_init_data_control(self);
+}
+
+static void client_detach_wayland(struct wayvnc_client* self)
+{
+	self->seat = NULL;
+
+	if (self->keyboard.virtual_keyboard) {
+		zwp_virtual_keyboard_v1_destroy(
+				self->keyboard.virtual_keyboard);
+		keyboard_destroy(&self->keyboard);
+	}
+	self->keyboard.virtual_keyboard = NULL;
+
+	if (self->pointer.pointer)
+		pointer_destroy(&self->pointer);
+	self->pointer.pointer = NULL;
+
+	if (self->data_control.manager)
+		data_control_destroy(&self->data_control);
+	self->data_control.manager = NULL;
+}
+
 static unsigned next_client_id = 1;
 
 static struct wayvnc_client* client_create(struct wayvnc* wayvnc,
@@ -1074,10 +1258,10 @@ static struct wayvnc_client* client_create(struct wayvnc* wayvnc,
 	self->nvnc_client = nvnc_client;
 
 	self->id = next_client_id++;
-	client_init_seat(self);
-	client_init_keyboard(self);
-	client_init_pointer(self);
-	client_init_data_control(self);
+
+	if (wayvnc->display) {
+		client_init_wayland(self);
+	}
 
 	return self;
 }
@@ -1088,6 +1272,12 @@ static void client_destroy(void* obj)
 	struct nvnc* nvnc = nvnc_client_get_server(self->nvnc_client);
 	struct wayvnc* wayvnc = nvnc_get_userdata(nvnc);
 
+	if (self == wayvnc->master_layout_client)
+		wayvnc->master_layout_client = NULL;
+
+	if (self->transient_seat)
+		ext_transient_seat_v1_destroy(self->transient_seat);
+
 	if (self->seat)
 		self->seat->occupancy--;
 
@@ -1095,16 +1285,15 @@ static void client_destroy(void* obj)
 	nvnc_log(NVNC_LOG_DEBUG, "Client disconnected, new client count: %d",
 			wayvnc->nr_clients);
 
-	struct ctl_server_client_info info = {
-		.id = self->id,
-		.hostname = nvnc_client_get_hostname(self->nvnc_client),
-		.username = nvnc_client_get_auth_username(self->nvnc_client),
-		.seat = self->seat ? self->seat->name : NULL,
-	};
+	if (wayvnc->ctl) {
+		struct ctl_server_client_info info = {};
+		compose_client_info(self, &info);
 
-	ctl_server_event_disconnected(wayvnc->ctl, &info, wayvnc->nr_clients);
+		ctl_server_event_disconnected(wayvnc->ctl, &info,
+				wayvnc->nr_clients);
+	}
 
-	if (wayvnc->nr_clients == 0) {
+	if (wayvnc->nr_clients == 0 && wayvnc->display) {
 		nvnc_log(NVNC_LOG_INFO, "Stopping screen capture");
 		screencopy_stop(&wayvnc->screencopy);
 		stop_performance_ticker(wayvnc);
@@ -1125,6 +1314,13 @@ static void client_destroy(void* obj)
 	free(self);
 }
 
+static void handle_first_client(struct wayvnc* self)
+{
+	nvnc_log(NVNC_LOG_INFO, "Starting screen capture");
+	start_performance_ticker(self);
+	wayvnc_start_capture_immediate(self);
+}
+
 static void on_nvnc_client_new(struct nvnc_client* client)
 {
 	struct nvnc* nvnc = nvnc_client_get_server(client);
@@ -1134,21 +1330,15 @@ static void on_nvnc_client_new(struct nvnc_client* client)
 	assert(wayvnc_client);
 	nvnc_set_userdata(client, wayvnc_client, client_destroy);
 
-	if (self->nr_clients == 0) {
-		nvnc_log(NVNC_LOG_INFO, "Starting screen capture");
-		start_performance_ticker(self);
-		wayvnc_start_capture_immediate(self);
+	if (self->nr_clients == 0 && self->display) {
+		handle_first_client(self);
 	}
 	self->nr_clients++;
 	nvnc_log(NVNC_LOG_DEBUG, "Client connected, new client count: %d",
 			self->nr_clients);
 
-	struct ctl_server_client_info info = {
-		.id = wayvnc_client->id,
-		.hostname = nvnc_client_get_hostname(client),
-		.username = nvnc_client_get_auth_username(client),
-		.seat = wayvnc_client->seat ? wayvnc_client->seat->name : NULL,
-	};
+	struct ctl_server_client_info info = {};
+	compose_client_info(wayvnc_client, &info);
 
 	ctl_server_event_connected(self->ctl, &info, self->nr_clients);
 }
@@ -1193,6 +1383,51 @@ static void client_init_pointer(struct wayvnc_client* self)
 	}
 }
 
+static void handle_transient_seat_ready(void* data,
+		struct ext_transient_seat_v1* transient_seat,
+		uint32_t global_name)
+{
+	(void)transient_seat;
+
+	struct wayvnc_client* client = data;
+	struct wayvnc* wayvnc = client->server;
+
+	struct seat* seat = seat_find_by_id(&wayvnc->seats, global_name);
+	assert(seat);
+
+	client->seat = seat;
+}
+
+static void handle_transient_seat_denied(void* data,
+		struct ext_transient_seat_v1* transient_seat)
+{
+	(void)data;
+	(void)transient_seat;
+
+	// TODO: Something more graceful perhaps?
+	nvnc_log(NVNC_LOG_PANIC, "Transient seat denied");
+}
+
+static void client_init_transient_seat(struct wayvnc_client* self)
+{
+	struct wayvnc* wayvnc = self->server;
+
+	self->transient_seat =
+		ext_transient_seat_manager_v1_create(wayvnc->transient_seat_manager);
+
+	static const struct ext_transient_seat_v1_listener listener = {
+		.ready = handle_transient_seat_ready,
+		.denied = handle_transient_seat_denied,
+	};
+	ext_transient_seat_v1_add_listener(self->transient_seat, &listener,
+			self);
+
+	// TODO: Make this asynchronous
+	wl_display_roundtrip(wayvnc->display);
+
+	assert(self->seat);
+}
+
 static void client_init_seat(struct wayvnc_client* self)
 {
 	struct wayvnc* wayvnc = self->server;
@@ -1202,6 +1437,8 @@ static void client_init_seat(struct wayvnc_client* self)
 
 	if (wayvnc->selected_seat) {
 		self->seat = wayvnc->selected_seat;
+	} else if (wayvnc->use_transient_seat) {
+		client_init_transient_seat(self);
 	} else {
 		self->seat = seat_find_unoccupied(&wayvnc->seats);
 		if (!self->seat) {
@@ -1322,6 +1559,113 @@ void switch_to_prev_output(struct wayvnc* self)
 	switch_to_output(self, prev);
 }
 
+static char intercepted_error[256];
+
+static void intercept_cmd_error(const struct nvnc_log_data* meta,
+		const char* message)
+{
+	if (meta->level != NVNC_LOG_ERROR) {
+		nvnc_default_logger(meta, message);
+		return;
+	}
+
+	struct nvnc_log_data meta_override = *meta;
+	meta_override.level = NVNC_LOG_DEBUG;
+	nvnc_default_logger(&meta_override, message);
+
+	size_t len = strlen(intercepted_error);
+	if (len != 0 && len < sizeof(intercepted_error) - 2)
+		intercepted_error[len++] = '\n';
+
+	strlcpy(intercepted_error + len, message,
+			sizeof(intercepted_error) - len);
+}
+
+static struct cmd_response* on_attach(struct ctl* ctl, const char* display)
+{
+	struct wayvnc* self = ctl_server_userdata(ctl);
+	assert(self);
+
+	memset(intercepted_error, 0, sizeof(intercepted_error));
+	nvnc_set_log_fn_thread_local(intercept_cmd_error);
+
+	// TODO: Add optional output argument
+	bool ok = wayland_attach(self, display, NULL);
+
+	nvnc_set_log_fn_thread_local(NULL);
+
+	return ok ? cmd_ok() : cmd_failed("%s", intercepted_error);
+}
+
+static bool wayland_attach(struct wayvnc* self, const char* display,
+		const char* output)
+{
+	if (self->display) {
+		wayland_detach(self);
+	}
+
+	nvnc_log(NVNC_LOG_DEBUG, "Attaching to %s", display);
+
+	if (init_wayland(self, display) < 0) {
+		return false;
+	}
+
+	struct output* out;
+	if (output) {
+		out = output_find_by_name(&self->outputs, output);
+		if (!out) {
+			nvnc_log(NVNC_LOG_ERROR, "No such output: %s", output);
+			wayland_detach(self);
+			return false;
+		}
+	} else {
+		out = output_first(&self->outputs);
+		if (!out) {
+			nvnc_log(NVNC_LOG_ERROR, "No output available");
+			wayland_detach(self);
+			return false;
+		}
+	}
+
+	screencopy_init(&self->screencopy);
+	if (!self->screencopy.manager) {
+		nvnc_log(NVNC_LOG_ERROR, "Attached display does not implement wlr-screencopy-v1");
+		wayland_detach(self);
+		return false;
+	}
+
+	set_selected_output(self, out);
+
+	struct nvnc_client* nvnc_client;
+	for (nvnc_client = nvnc_client_first(self->nvnc); nvnc_client;
+			nvnc_client = nvnc_client_next(nvnc_client)) {
+		struct wayvnc_client* client = nvnc_get_userdata(nvnc_client);
+		client_init_wayland(client);
+	}
+
+	nvnc_log(NVNC_LOG_INFO, "Attached to %s", display);
+
+	if (self->nr_clients > 0) {
+		handle_first_client(self);
+	}
+
+	return true;
+}
+
+static struct cmd_response* on_detach(struct ctl* ctl)
+{
+	struct wayvnc* self = ctl_server_userdata(ctl);
+	assert(self);
+
+	if (!self->display) {
+		return cmd_failed("Not attached!");
+	}
+
+	wayland_detach(self);
+	nvnc_log(NVNC_LOG_INFO, "Detached from wayland server");
+	return cmd_ok();
+}
+
 static int log_level_from_string(const char* str)
 {
 	if (0 == strcmp(str, "quiet")) return NVNC_LOG_PANIC;
@@ -1352,6 +1696,7 @@ int main(int argc, char* argv[])
 	int port = 0;
 	bool use_unix_socket = false;
 	bool use_websocket = false;
+	bool start_detached = false;
 
 	const char* output_name = NULL;
 	const char* seat_name = NULL;
@@ -1362,6 +1707,7 @@ int main(int argc, char* argv[])
 	bool show_performance = false;
 	int max_rate = 30;
 	bool disable_input = false;
+	bool use_transient_seat = false;
 
 	int drm_fd MAYBE_UNUSED = -1;
 
@@ -1386,6 +1732,8 @@ int main(int argc, char* argv[])
 		  "Select seat by name." },
 		{ 'S', "socket", "<path>",
 		  "Control socket path." },
+		{ 't', "transient-seat", NULL,
+		  "Use transient seat." },
 		{ 'r', "render-cursor", NULL,
 		  "Enable overlay cursor rendering." },
 		{ 'f', "max-fps", "<fps>",
@@ -1397,6 +1745,8 @@ int main(int argc, char* argv[])
 		  "Create unix domain socket." },
 		{ 'd', "disable-input", NULL,
 		  "Disable all remote input." },
+		{ 'D', "detached", NULL,
+		  "Start detached from a compositor." },
 		{ 'V', "version", NULL,
 		  "Show version info." },
 		{ 'v', "verbose", NULL,
@@ -1441,6 +1791,11 @@ int main(int argc, char* argv[])
 	log_level = log_level_from_string(
 			option_parser_get_value(&option_parser, "log-level"));
 	max_rate = atoi(option_parser_get_value(&option_parser, "max-fps"));
+	use_transient_seat = !!option_parser_get_value(&option_parser,
+				"transient-seat");
+	start_detached = !!option_parser_get_value(&option_parser, "detached");
+
+	self.start_detached = start_detached;
 
 	keyboard_options = option_parser_get_value(&option_parser, "keyboard");
 	if (keyboard_options)
@@ -1465,6 +1820,16 @@ int main(int argc, char* argv[])
 		return 1;
 	}
 
+	if (use_transient_seat && disable_input) {
+		nvnc_log(NVNC_LOG_ERROR, "transient-seat and disable-input are conflicting options");
+		return 1;
+	}
+
+	if (seat_name && use_transient_seat) {
+		nvnc_log(NVNC_LOG_ERROR, "transient seat and seat are conflicting options");
+		return 1;
+	}
+
 	errno = 0;
 	int cfg_rc = cfg_load(&self.cfg, cfg_file);
 	if (cfg_rc != 0 && (cfg_file || errno != ENOENT)) {
@@ -1490,38 +1855,51 @@ int main(int argc, char* argv[])
 	if (!port) port = DEFAULT_PORT;
 
 	self.disable_input = disable_input;
+	self.use_transient_seat = use_transient_seat;
 
-	if (init_wayland(&self) < 0) {
-		nvnc_log(NVNC_LOG_ERROR, "Failed to initialise wayland");
-		return 1;
+	struct aml* aml = aml_new();
+	if (!aml)
+		goto failure;
+
+	aml_set_default(aml);
+
+	if (init_main_loop(&self) < 0)
+		goto failure;
+
+	if (!start_detached) {
+		if (init_wayland(&self, NULL) < 0) {
+			nvnc_log(NVNC_LOG_ERROR, "Failed to initialise wayland");
+			goto wayland_failure;
+		}
+
+		struct output* out;
+		if (output_name) {
+			out = output_find_by_name(&self.outputs, output_name);
+			if (!out) {
+				nvnc_log(NVNC_LOG_ERROR, "No such output");
+				goto wayland_failure;
+			}
+		} else {
+			out = output_first(&self.outputs);
+			if (!out) {
+				nvnc_log(NVNC_LOG_ERROR, "No output found");
+				goto wayland_failure;
+			}
+		}
+		set_selected_output(&self, out);
+
+		struct seat* seat = NULL;
+		if (seat_name) {
+			seat = seat_find_by_name(&self.seats, seat_name);
+			if (!seat) {
+				nvnc_log(NVNC_LOG_ERROR, "No such seat");
+				goto wayland_failure;
+			}
+		}
+
+		self.selected_seat = seat;
 	}
 
-	struct output* out;
-	if (output_name) {
-		out = output_find_by_name(&self.outputs, output_name);
-		if (!out) {
-			nvnc_log(NVNC_LOG_ERROR, "No such output");
-			goto failure;
-		}
-	} else {
-		out = output_first(&self.outputs);
-		if (!out) {
-			nvnc_log(NVNC_LOG_ERROR, "No output found");
-			goto failure;
-		}
-	}
-	set_selected_output(&self, out);
-
-	struct seat* seat = NULL;
-	if (seat_name) {
-		seat = seat_find_by_name(&self.seats, seat_name);
-		if (!seat) {
-			nvnc_log(NVNC_LOG_ERROR, "No such seat");
-			goto failure;
-		}
-	}
-
-	self.selected_seat = seat;
 	self.screencopy.rate_limit = max_rate;
 	self.screencopy.enable_linux_dmabuf = enable_gpu_features;
 
@@ -1534,15 +1912,6 @@ int main(int argc, char* argv[])
 	if (aml_unstable_abi_version != AML_UNSTABLE_API)
 		nvnc_log(NVNC_LOG_PANIC, "libaml is incompatible with this build of wayvnc!");
 
-	struct aml* aml = aml_new();
-	if (!aml)
-		goto main_loop_failure;
-
-	aml_set_default(aml);
-
-	if (init_main_loop(&self) < 0)
-		goto main_loop_failure;
-
 	enum socket_type socket_type = SOCKET_TYPE_TCP;
 	if (use_unix_socket)
 		socket_type = SOCKET_TYPE_UNIX;
@@ -1552,12 +1921,14 @@ int main(int argc, char* argv[])
 	if (init_nvnc(&self, address, port, socket_type) < 0)
 		goto nvnc_failure;
 
-	if (self.screencopy.manager)
-		screencopy_init(&self.screencopy);
+	if (!start_detached) {
+		if (self.screencopy.manager)
+			screencopy_init(&self.screencopy);
 
-	if (!self.screencopy.manager) {
-		nvnc_log(NVNC_LOG_ERROR, "screencopy is not supported by compositor");
-		goto capture_failure;
+		if (!self.screencopy.manager) {
+			nvnc_log(NVNC_LOG_ERROR, "screencopy is not supported by compositor");
+			goto capture_failure;
+		}
 	}
 
 	self.screencopy.overlay_cursor = overlay_cursor;
@@ -1568,6 +1939,8 @@ int main(int argc, char* argv[])
 
 	const struct ctl_server_actions ctl_actions = {
 		.userdata = &self,
+		.on_attach = on_attach,
+		.on_detach = on_detach,
 		.on_output_cycle = on_output_cycle,
 		.on_output_switch = on_output_switch,
 		.client_next = client_next,
@@ -1580,32 +1953,35 @@ int main(int argc, char* argv[])
 	if (!self.ctl)
 		goto ctl_server_failure;
 
-	wl_display_dispatch_pending(self.display);
+	if (self.display)
+		wl_display_dispatch_pending(self.display);
 
 	while (!self.do_exit) {
-		wl_display_flush(self.display);
+		if (self.display)
+			wl_display_flush(self.display);
+
 		aml_poll(aml, -1);
 		aml_dispatch(aml);
 	}
 
 	nvnc_log(NVNC_LOG_INFO, "Exiting...");
 
-	screencopy_stop(&self.screencopy);
+	if (self.display)
+		screencopy_stop(&self.screencopy);
+
 	ctl_server_destroy(self.ctl);
+	self.ctl = NULL;
 
 	nvnc_display_unref(self.nvnc_display);
 	nvnc_close(self.nvnc);
-	if (zwp_linux_dmabuf)
-		zwp_linux_dmabuf_v1_destroy(zwp_linux_dmabuf);
-	if (self.screencopy.manager)
-		screencopy_destroy(&self.screencopy);
+	self.nvnc = NULL;
+	wayvnc_destroy(&self);
 #ifdef ENABLE_SCREENCOPY_DMABUF
 	if (gbm_device) {
 		gbm_device_destroy(gbm_device);
 		close(drm_fd);
 	}
 #endif
-	wayvnc_destroy(&self);
 	aml_unref(aml);
 
 	return 0;
@@ -1615,15 +1991,16 @@ capture_failure:
 	nvnc_display_unref(self.nvnc_display);
 	nvnc_close(self.nvnc);
 nvnc_failure:
+wayland_failure:
 	aml_unref(aml);
-main_loop_failure:
 failure:
+	self.nvnc = NULL;
+	wayvnc_destroy(&self);
 #ifdef ENABLE_SCREENCOPY_DMABUF
 	if (gbm_device)
 		gbm_device_destroy(gbm_device);
 	if (drm_fd >= 0)
 		close(drm_fd);
 #endif
-	wayvnc_destroy(&self);
 	return 1;
 }
